{ applications.pp: A basic framework for representing a single process and its
    main thread of execution in an object-oriented fashion.

  Copyright (C) 2016-2019 Michael Malicoat <michael@circusmachina.com>.

  This file is part of ASE, the Abandonded Sessions Eliminator.

  This program is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation; either version 3 of the License or, at your option, any later
  version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABLITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.

  You should have received a copy of the GNU General Public License along with
  this program.  If not, see <http://www.gnu.org/licenses/>.
}
unit applications;

{$mode objfpc}{$H+}

interface

uses
  Classes, Contnrs, CustApp, SysUtils;

{ This class can be used to represent a process and its main thread of execution
  in an object-oriented fashion.  It extends @code(TCustomApplication) to
  provide several utility functions, such as the ability to expand environment
  variables contained within a string
  (@link(AnApplication.expandEnvironmentVariablesIn)).
}
type AnApplication = class(TCustomApplication)
  protected
    // The application exit code
    myExitCode: ptrInt;
    { Stores the environment variables available to the process }
    MyEnvironment: TFPStringHashTable;
    { Stores the command-line parameters available to the process }
    MyParameters: TStrings;
    { Stores the version information available to the process }
    MyVersionInfo: TFPStringHashTable;

    { Parse any and all environment variables made available to the process.
      This method separates the name of the variable from its value and stores
      both in @link(AnApplication.Environment).  It is called automatically by
      @link(AnApplication.init).
    }
    procedure parseEnvironment; virtual;

    { Parse the command line arguments made available to the process.  This
      method collects all parameters except the zeroth one (the name of the
      application being executed) into @link(AnApplication.Parameters).  It is
      called automatically by @link(AnApplication.init).
    }
    procedure storeCommandLine; virtual;

    // Handle the occurrence of a command-line argument
    function handleCommandLineArgument(const argument: string;
      const parameter: string): boolean; virtual;

    // Process command-line arguments
    function processCommandLine: longint; virtual;

    // Execute the application
    procedure doRun; override;

  public
    // Constructor
    constructor create(TheOwner: TComponent); override;

    { Initializer }
    function init: boolean; virtual;

    { Destroy the application instance.

      This method is called automatically by @code(TObject.free) when it is
      called on an instance of @link(AnApplication).  It frees
      @link(AnApplication.Environment) and @link(AnApplication.Parameters)
      before calling the inherited routine.
    }
    destructor destroy; override;

    { Execute the application.

      This method should be overridden by descendant classes.  In the base
      implementation, it does nothing.

      @return(A value indicating whether the application executed successfully.
      This value may be passed back to the underlying operating system.
      Generally, a return value of zero (0) means no errors occurred.  A
      value greater than zero indicates some kind of error condition, with the
      exact meaning dependent on the application.  A value less than zero
      indicates a condition which is not an error, but which can nonetheless
      prevent normal execution from proceeding.)
    }
    function run: ptrInt; virtual;

    { Suspend execution of the application for at least @code(count)
      milliseconds.  Depending on the underlying operating system, the actual
      amount of time that the process is suspended may be greater than
      @code(count).

      This method simply calls @code(SysUtils.sleep).  It is provided for
      convenience.
    }
    procedure sleep(const count: longword); virtual;

    { Retrieve the absolute path of the executable; this does NOT include the
      name of the executable file itself.  To obtain the name of the executable
      file without its path, call @link(AnApplication.executableName).
    }
    function executablePath: string; virtual;

    { Retrieve the name of the executable file that contains the application.
      This is the name of the file without any path information; to obtain the
      full path to the executable, call @link(AnApplication.executablePath).

      To obtain the name of the executable file @italic(without) its extension,
      set @code(withoutExtension) to @true.
    }
    function executableName(const withoutExtension: boolean = false): string;
      virtual;

    { Get the current working directory of the appliction.

      This method simply calls @code(SysUtils.getCurrentDir) and returns the
      result.  It is provided for convenience.
    }
    function currentDirectory: string; virtual;

    { Set the current working directory of the application.

      This method simply calls @code(SysUtils.setCurrentDir).  It is provided
      for convenience.

      @return(The previous value of @link(AnApplication.currentDirectory).)
    }
    function setCurrentDirectory(const newPath: string): string; virtual;

    { Get the name of the directory where temporary files may be stored.
      Depending on the underlying operating system, this directory may be
      specific to the user that is running the application; to obtain the name
      of a temporary directory that is accessible to all users, set
      @code(forAllUsers) to @true.

      This method simply calls @code(SysUtils.getTempDir) and returns the
      result.  Note that the resulting string will contain a path delimiter.
    }
    function temporaryDirectory(const forAllUsers: boolean = false): string;
      virtual;

    { Get the name of the directory where application configuration files may
      be stored.  Depending on the underlying operating system, this directory
      may be specific to the user that is running the application; to obtain the
      name of a temporary directory that is accessible to all users, set
      @code(forAllUsers) to @true.

      This method simply calls @code(SysUtils.getAppConfigDir) and returns the
      result.  Note that the resulting string will contain a path delimiter.
    }
    function configurationDirectory(const forAllUsers: boolean = false): string;
      virtual;

    { Get the name of the directory where data for the current user is kept.

      This method simply calls @code(SysUtils.getUserDir) and returns the
      result.  Note that the resulting string will contain a path delimiter.
    }
    function userDirectory: string; virtual;

    { Expand a string that contains environment variable references }
    function expandEnvironmentVariablesIn(const thisString: string): string;
      virtual;

    // Determine the size of the named file
    function fileSize(const thisFileName: string): int64; virtual;

    // Properties --------------------------------------------------------------
    { Retrieve a reference to the dictionary of environment variables available
      to the process.  This reference should NOT be freed by the caller; that
      will be done when the @link(AnApplication) instance is destroyed.

      The dictionary reference returned by this method will contain any and all
      environment variables made available to the process.
    }
    function Environment: TFPStringHashTable; virtual;

    { Retrieve a reference to the dictionary of command-line parameters
      available to the process.  This reference should NOT be freed by the
      caller; that will be done when the @link(AnApplication) instance is
      destroyed.

      The dictionary reference returned by this method will contain any and all
      parameters passed to the process.  It will NOT contain the name of the
      application itself (the "zeroth" argument); this is instead returned by
      @link(AnApplication.executablePath).
    }
    function Parameters: TStrings; virtual;

    { Retrieve a reference to the dictionary of version information available
      to the process.  This reference should NOT be freed by the caller; that
      will be done with the @link(AnApplication) instance is destroyed.

      The dictionary reference returned by this method will contain any and all
      version information key/value pairs that were compiled as a resource into
      the object file which represents the main executable of the current
      process.  This dictionary is cached, as these values will not change
      throughout the lifetime of the program.  The dictionary is constructed
      the first time this method is called.
    }
    function VersionInfo: TFPStringHashTable; virtual;

    // Retrieve the application exit code
    function exitCode: ptrInt; virtual;
end;

const
  // Error message used when the application could not be initialized
  appErrorAppInit = 'failed to initialize the application';

  { This constant is used as the return value of
    @link(AnApplication.run) to indicate that execution completed successfully.

    The general convention is that
    @link(AnApplication.run) should return zero (@code(0)) when
    execution proceeds sucessfully.  It should return a value greater than
    zero to indicate an error condition.  It should return a value less than
    zero to indicate something which is not an error but which prevented
    normal execution from proceeding.
  }
  APP_RUN_COMPLETE = 0;

  { This constant is used as the return value of
    @link(AnApplication.run) to indicate that the return value
    is not necessarily an error, but which nonetheless prevented normal
    execution from proceeding.  Non-fatal errors should be flagged this way,
    as should special-case execution scenarios.

    The general convention is that
    @link(AnApplication.run) should return zero (@code(0)) when
    execution proceeds sucessfully.  It should return a value greater than
    zero to indicate an error condition.  It should return a value less than
    zero to indicate something which is not necessarily an error but which
    prevented normal execution from proceeding.
  }
  APP_RUN_INCOMPLETE = -1;

  { The delimiter used to indicate a variable reference }
  appDelimVariable = '%';
  { The delimiter used to separate an environment variable key from its value }
  appDelimKeyValue = '=';

  // The characters that can begin a command-line command
  appDelimCommandLineCommand = ['-', '/'];
  // The character that delimimts a command-line command parameter
  appDelimCommandLineParam = '=';

implementation

uses
  StrUtils, VersionInfo;

const
  { This allows us to deal with the ambiguities of trying to specify special
    characters on the command line.  Because some characters are recognized by
    the command processor, it becomes difficult to include them in arguments
    that should be passed to the shortcut target; thus, we allow the user to
    provide a recognized HTML entity name instead, which we then convert to its
    associated character when saving the shortcut.
  }
  entityStrings: array[0..30] of string = (
    '&excl;',
    '&quot;',
    '&num;',
    '&dollar;',
    '&percnt;',
    '&amp;',
    '&apos;',
    '&lpar;',
    '&rpar;',
    '&ast;',
    '&plus;',
    '&comma;',
    '&period;',
    '&sol;',
    '&colon;',
    '&semi;',
    '&lt;',
    '&gt;',
    '&quest;',
    '&commat;',
    '&lbrack;',
    '&bsol;',
    '&rbrack;',
    '&Hat;',
    '&grave;',
    '&lbrace;',
    '&vert;',
    '&rbrace;',
    '&nbsp;',
    '&dash;',
    '&hyphen;'
  );

  entityReplacements: array[0..30] of string = (
    '!',
    '"',
    '#',
    '$',
    '%',
    '&',
    '''',
    '(',
    ')',
    '*',
    '+',
    ',',
    '.',
    '/',
    ':',
    ';',
    '<',
    '>',
    '?',
    '@',
    '[',
    '\',
    ']',
    '^',
    '`',
    '{',
    '|',
    '}',
    ' ',
    '-',
    '-'
  );

// AnApplication ***************************************************************
// Constructor
constructor AnApplication.create(TheOwner: TComponent);

begin
  inherited create(TheOwner);

  if Self.init = false then
    raise Exception.create(appErrorAppInit);

  stopOnException := true;
end;

// Initializer
function AnApplication.init: boolean;

begin
  result := false;

  // Construct our environment dictionary and command-line parameters list
  MyEnvironment := TFPStringHashTable.create;
  MyParameters := TStringList.create;
  MyVersionInfo := nil;

  if (MyEnvironment = nil) or (MyParameters = nil) then
    exit;

  // Parse the environment
  Self.parseEnvironment;
  // Get the command-line
  Self.storeCommandLine;

  result := true;
end;

// Destructor
destructor AnApplication.destroy;

begin
  // Free our environment dictionary and list of command-line parameters
  if MyEnvironment <> nil then
    MyEnvironment.free;

  if MyParameters <> nil then
    MyParameters.free;

  // Free the version info dictionary
  if MyVersionInfo <> nil then
    MyVersionInfo.free;

  inherited destroy;
end;

// Parse the environment
procedure AnApplication.parseEnvironment;

var
  // Loop counter
  i: sizeInt;
  // Stores the value of an environment variable
  variableValue: string;
  // Stores the name of an environment variable
  variableName: string;
  // Stores the position of the delimiter between names and values
  delimiterPos: sizeInt;

begin
  // Loop through each variable
  for i := 1 to SysUtils.getEnvironmentVariableCount do
  begin
    variableValue := SysUtils.getEnvironmentString(i);
    { On Windows, environment strings can begin with an equals sign, which will
      confuse our parser, since it expects the first equals sign to be the
      delimiter.  So we use the LAST position of the equals sign as the place
      to separate our variable name from its value.
    }
    delimiterPos := RPos(appDelimKeyValue, variableValue);

    if delimiterPos > 0 then
    begin
      variableName := copy(variableValue, 1, delimiterPos - sizeof(char));
      variableValue := copy(variableValue,
        delimiterPos + System.length(appDelimKeyValue),
        System.length(variableValue));

      // Add the variable name and value to our environment dictionary
    {$IFDEF WINDOWS}
      // Environment variable names are not case-sensitive on Windows
      variableName := ansiLowerCase(variableName);
    {$ENDIF WINDOWS}

      // Add the variable name and value to our environment dictionary
      MyEnvironment.add(variableName, variableValue);
      variableName := '';
      variableValue := '';
    end;
  end;
end;

// Get the list of command-line parameters passed to the application
procedure AnApplication.storeCommandLine;

var
  // Loop counter
  i:  sizeInt;

begin
  for i := 1 to System.paramCount do
    { We don't explicitly specify System.paramStr() here because the ObjPas unit
      overrides the definition to provide a version that accepts command lines
      greater than 255 characters
    }
    MyParameters.add(paramStr(i));
end;

// Handle the occurrence of a command-line argument
function AnApplication.handleCommandLineArgument(const argument: string;
  const parameter: string): boolean;

begin
  // Nothing to do here; always fail until overridden by a descendant
  result := false;
end;

// Process the command line
function AnApplication.processCommandLine: longint;

var
  // Loop counter
  i: longint;
  // Refers to a single parameter item
  command, param: string;
  // Refers to the position of a command-line parameter
  paramPos: longint;

begin
  result := -1;

  // No arguments passed to the executable
  if Self.Parameters.count = 0 then
  begin
    result := 0;
    exit;
  end;

  i := 0;
  while i < Self.Parameters.count do
  begin
    command := '';
    param := Self.Parameters[i];

    // Replace any HTML entities in the string
    param := StrUtils.stringsReplace(param, entityStrings, entityReplacements,
      [rfReplaceAll]);

    // Check for a command line command
    if param[1] in appDelimCommandLineCommand then
    begin
      // Commands are case-insensitive
      command := lowerCase(param);

      // Strip the command delimiter character from the string
      if System.length(command) > 1 then
        repeat
          command := System.copy(command, 2, System.length(command));
        until (command[1] in appDelimCommandLineCommand = false) or
          (System.length(command) <= 1);

      // The command consisted only of command delimiters; stop processing
      if (System.length(command) = 1) and
        (command[1] in appDelimCommandLineCommand) then
        break;

      // Check for a parameter passed with the command
      paramPos := pos(appDelimCommandLineParam, param);
      if paramPos > 0 then
      begin
        // Split the command into a command + parameter
        param :=
          System.copy(param,
            paramPos + System.length(appDelimCommandLineParam),
            System.length(param));

        System.setLength(command, paramPos - (sizeof(char) shl 1));
      end

      else
        param := '';
    end;

    if Self.handleCommandLineArgument(command, param) = false then
      exit;

    inc(i);
  end;

  // Return the number of parameters processed
  result := i;
end;

// Execute the main body of the application
procedure AnApplication.doRun;

begin
  myExitCode := Self.run;

  // stop program loop
  terminate;
end;

// Run the application
function AnApplication.run: ptrInt;

begin
  result := Self.processCommandLine;

  if result < 0 then
    exit;

  { Descendant classes should override this. }
  result := 0;
end;

// Sleep for at least the specified number of milliseconds
procedure AnApplication.sleep(const count: longword);

begin
  SysUtils.sleep(count);
end;

// Retrieve the full path to the executable
function AnApplication.executablePath: string;

begin
  // Get the full path to the executable
  result := SysUtils.expandFileName(System.paramStr(0));
  // Now, remove the name of the executable file itself
  result := SysUtils.extractFilePath(result);

  if result = '' then
    result := Self.currentDirectory;
end;

// Retrieve the name of the executable
function AnApplication.executableName(const withoutExtension: boolean = false): string;

begin
  result := SysUtils.extractFileName(System.paramStr(0));

  // Strip the extension, if desired
  if withoutExtension then
    result := SysUtils.changeFileExt(result, '');
end;

// Get the current working directory
function AnApplication.currentDirectory: string;

begin
  result := SysUtils.getCurrentDir;
end;

// Set the current working directory
function AnApplication.setCurrentDirectory(const newPath: string): string;

begin
  // Get the current directory
  result := SysUtils.getCurrentDir;

  // Set the new directory
  SysUtils.setCurrentDir(newPath);
end;

// Get the system-defined temporary file directory
function AnApplication.temporaryDirectory(const forAllUsers: boolean = false): string;

begin
  result := SysUtils.getTempDir(forAllUsers);
end;

// Get the system-defined application configuration directory
function AnApplication.configurationDirectory(const forAllUsers: boolean = false): string;

begin
  result := SysUtils.getAppConfigDir(forAllUsers);
end;

// Get the system-defined user data directory for the current user
function AnApplication.userDirectory: string;

begin
  result := SysUtils.getUserDir;
end;

// Expand a string that contains variable references
{ This could probably be more efficient, but it's cross-platform }
function AnApplication.expandEnvironmentVariablesIn(const thisString: string): string;

var
  // Loop counter
  i: sizeInt;
  // Stores the length of the string
  stringLength: sizeInt;
  // Stores a character from the string
  ch: string;
  // Stores a token from the string
  token: string;
  // Stores a variable name
  varName: string;

begin
  i := 1;
  token := '';
  result := '';
  stringLength := System.length(thisString);

  while i <= stringLength do
  begin
    ch := thisString[i];

    if ch = appDelimVariable then
    begin
      // Are we at the end of an environment variable spec?
      if (token <> '') and
        (token[1] = appDelimVariable) then
      begin
        // Get the value of the named variable without the opening '%'
        varName := copy(token, System.length(appDelimVariable) + sizeof(char),
          System.length(token));

      {$IFDEF WINDOWS}
        // Environment variable names are not case-sensitive on Windows
        varName := ansiLowerCase(varName);
      {$ENDIF WINDOWS}

        result := result + MyEnvironment[varName];
        token := '';
      end

      // Two % characters in a row
      else if (i < (stringLength - 1)) and
        (thisString[i + 1] = appDelimVariable) then
      begin
        token := token + appDelimVariable;
        inc(i);
      end

      // Otherwise, the possible beginning of a new environment variable name
      else begin
        result := result + token;
        token := ch;
      end;
    end

    // Environment variable names cannot have spaces
    else if ch = ' ' then
    begin
      result := result + token + ch;
      token := '';
    end

    else
      token := token + ch;

    inc(i);
  end;

  // Save the final token
  result := result + token;
end;

// Determine the size of the named file
function AnApplication.fileSize(const thisFileName: string): int64;

var
  // A file stream
  ThisFile: TFileStream;

begin
  result := 0;

  if (thisFileName = '') or (SysUtils.fileExists(thisFileName) = false) then
    exit;

  try
    ThisFile := TFileStream.create(thisFileName, fmOpenRead or fmShareDenyWrite);
    result := ThisFile.size;

  finally
    ThisFile.free;
  end;
end;

// Properties ------------------------------------------------------------------
// Retrieve a reference to the dictionary of environment variables
function AnApplication.Environment: TFPStringHashTable;

begin
  Result := MyEnvironment;
end;

// Retrieve a reference to the list of command-line parameters
function AnApplication.Parameters: TStrings;

begin
  Result := MyParameters;
end;

// Retrieve the version information
function AnApplication.VersionInfo: TFPStringHashTable;

var
  // Refers to the version information resource for the current executable
  VersionResource: AFileVersionInfoResource;

begin
  if MyVersionInfo = nil then begin
    MyVersionInfo := TFPStringHashTable.create;

  {$IFNDEF WINDOWS}
    VersionResource := AnELFFileVersionInfoResource.create;
  {$ELSE}
    VersionResource := AWindowsFileVersionInfoResource.create;
  {$ENDIF}

    // Read the version information
    VersionResource.readInto(MyVersionInfo);
    VersionResource.free;
  end;

  Result := MyVersionInfo;
end;

// Retrieve the application exit code
function AnApplication.exitCode: ptrInt;

begin
  result := myExitCode;
end;

end.

