{ aseapplication.pp

  This unit defines the class used to represent the main thread of execution for
  ASE.

  This file is part of ASE, the Abandonded Sessions Eliminator.

  This program is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation; either version 3 of the License or, at your option, any later
  version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABLITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.

  You should have received a copy of the GNU General Public License along with
  this program.  If not, see <http://www.gnu.org/licenses/>.
}
unit aseapplication;

{$mode objfpc}{$H+}

interface

uses
  Applications, Classes, Contnrs, EventLog, WTSAPI32;

type AnASEApplication = class(AnApplication)
  protected
    // The application configuration
    MyConfiguration: TFPStringHashTable;
    // The logging mechanism used by the application
    MyLog: TEventLog;
    // The maximum number of seconds a session may be idle
    mySessionIdleMaximum: qword;
    // The list of idle sessions
    MyIdleSessions: TQueue;

    // Handle the occurrence of a command-line argument
    function handleCommandLineArgument(const argument: string;
      const parameter: string): boolean; override;

    // Read configuration
    function readConfiguration: boolean; virtual;

    // Begin logging
    function beginLogging: boolean; virtual;

    { Calculate the idle time for a given session.

      In spite of what is suggested by the MSDN documentation of
      @link(WTSINFOA WTSINFO), in practice it seems that the
      @link(WTSINFOA.LastInputType LastInputType) member is only updated when
      the user connects to a session.  This means the user with the active
      session will show as idle, too.

      As a result, this function uses the value of the
      @link(WTSINFOA.DisconnectTime DisconnectTime) member to determine how long
      a session has been disconnected -- however, practice has shown that this
      value is also updated for the active user session, even though it's still
      connected.  This function therefore ignores @code(thisSession) if its
      @code(WTSINFOA.State State) member is not
      @link(WTS_CONNECTSTATE_CLASS.WTSDisconnected WTSDisconnected).

      @param(thisSession a pointer to a link(WTSINFOA WTSINFO) structure that
                         was obtained by means of a call to
                         @link(WTSQuerySessionInfo).

        )

      @return(The number of seconds that @code(thisSession) has been idle,
        subject to the caveats listed above.  If @code(thisSession) is not
        disconnected, the function returns @code(0).
      )
    }
    function getIdleTimeSecondsFrom(const thisSession: PWTSINFO): qword;
      virtual;
    // Collect idle sessions
    function collectIdleSessions: sizeInt; virtual;

  public
    // Initializer
    function init: boolean; override;
    // Destructor
    destructor destroy; override;

    // Eject idle sessions
    function ejectIdleSessions: sizeInt; virtual;

    // Execute the main body of the application
    function run: ptrInt; override;

    // Print usage information to stdout
    procedure printUsage; virtual;

    // Log a note
    procedure logNoteMessage(const thisMessage: string); virtual;
    // Log a note
    procedure logNoteFmt(const fmt: string; const args: array of const);
      virtual;

    // Log an error message
    procedure logErrorMessage(const thisMessage: string); virtual;
    // Log an error message
    procedure logErrorFmt(const fmt: string; const args: array of const);
      virtual;
  end;

resourcestring
  // Configuration -------------------------------------------------------------
  // The application title; used when logging messages to the log
  aseApplicationTitle = 'ASE';

  { Used to construct the path to the configuration file if a patch is not
    specified on the command-line
  }
  aseFmtINIFilePath = '%%ProgramData%%\%0:s%0:s.ini';

  { The key paired with the setting that indicates the path and name of the INI
    file to use for configuration
  }
  aseKeyConfigFilePath = 'strcfgfilepath';

  { The key paired with the setting that indicates whether or not to log to the
    system log, if available
  }
  aseKeyLogToSyslog = 'boollogtosyslog';

  // By default, we log to the system log
  aseDefaultLogToSyslog = 'y';

  { The key paired with the setting that indicates the path and name of the file
    to use when logging to a file
  }
  aseKeyLogFilePath = 'strlogfilepath';

  { The key paired with the setting that indictes the maximum number of seconds
    a session may be idle before it is considered for ejection
  }
  aseKeyEjectIdleAfterSeconds = 'intejectidleafterseconds';

  // The default number of seconds before an idle session is ejected
  aseDefaultEjectIdleAfterSeconds = '3600';

  // Program usage information
  aseMsgUsageFmt =
    '%s v%s' + lineEnding +
    'Copyright (C) 2018-2019 Michael Malicoat <michael@circusmachina.com>.' + lineEnding +
    'This program is distributed WITHOUT ANY WARRANTY; without even the implied' + lineEnding +
    'warranty of MERCHANTABLITY or FITNESS FOR A PARTICULAR PURPOSE.' + lineEnding +
    lineEnding +
    'Usage: %s <options>' + lineEnding +
    lineEnding +
    '<options> may be one or more of the following:' + lineEnding +
    '  -?, --help              Print this usage information' + lineEnding +
    lineEnding +
    '  -cf, --config=<path>    Specify the path and file name of an INI file to use ' + lineEnding +
    '                          for program configuration' + lineEnding +
    lineEnding +
    '  -lf, --log=<path>       Specify the path and file name of a file to use for ' + lineEnding +
    '                          output of log information.  This option is mutually ' + lineEnding +
    '                          exclusive of the `syslog` option.' + lineEnding +
    lineEnding +
    '  -ls, --syslog           Indicate the program should log to the system log ' + lineEnding +
    '                          (the Windows Event Log) instead of to a file.  This ' + lineEnding +
    '                          option is mutually exclusive of the `log` option.' + lineEnding +
    '                          This option is enabled by default.' + lineEnding +
    lineEnding +
    '  -t, --timeout=<integer> Specify the number of seconds that a disconnected' + lineEnding +
    '                          user session is allowed to remain idle before it is' + lineEnding +
    '                          forcibly logged off the machine.  The default value ' + lineEnding +
    '                          of this option is "%s".' + lineEnding +
    lineEnding;

  // Error messages ------------------------------------------------------------
  // Error message used for an unrecognized command-line argument
  aseErrorUnrecognizedArgument = 'error: unknown command-line argument: "%s"';
  // Error message used when the configuration file is not found
  aseErrorMissingINI = 'error: missing configuration file: "%s"';

  // Error message used when readConfiguration() fails
  aseErrorReadingConfiguration = 'error: failed to configure application';

  // Format of the note used to log that one or more sessions were ejected
  aseNoteEjectIdleSessions = 'The following user sessions exceeded the ' +
    'idle time limit of %d second(s) and were logged off:' + lineEnding +
    lineEnding +
    'SESSION    USER NAME' + lineEnding +
    '---------- ------------------------------------------------' + lineEnding +
    '%s';

  // Format of each line in the above note, indicating an ejected session
  aseFmtIdleSessionUserName = '  %4d     %s\%s ' +
    '[idle for %d second(s)]' + lineEnding;

  // Format of the note used to log that no sessions were eligible for ejection
  aseNoteNoIdleSessions = 'There were no user sessions that exceeded the ' +
    'idle time limit of %d second(s)';

  // Error message used when collectIdleSessions() fails
  aseErrorCollectIdleSessions = 'Failed to enumerate available desktop ' +
    'sessions';

implementation

uses
  IniFiles, SysUtils, VersionInfo;

// Convert a string value to a boolean value
function boolFromString(const thisString: string): boolean;

begin
  result := false;

  if thisString = '' then
    exit;

  case thisString[1] of
    // 'on', 'off', 'On', 'Off', 'ON', 'OFF'
    'o', 'O':
      if System.length(thisString) > 1 then
        case thisString[2] of
          'n', 'N':
            result := true;
        end;

    // 't', 'T', 'true', 'True', or 'TRUE'
    't', 'T':
      result := true;

    // 'y', 'Y', 'yes', 'Yes', or 'YES'
    'y', 'Y':
      result := true;

    // Any non-zero positive numeric value
    '1'..'9':
      result := true;

    // Any non-zero negative numeric value
    '-':
      if System.length(thisString) > 1 then
        case thisString[2] of
          '1'..'9':
            result := true;
        end;
  end;
end;

// AnASEApplication ************************************************************
// Initializer
function AnASEApplication.init: boolean;

begin
  result := inherited init;

  if result then
  begin
    result := false;

    Self.title := aseApplicationTitle;
    MyConfiguration := TFPStringHashTable.create;
    if MyConfiguration = nil then
      exit;

    { Create our configuration keys with their default values; these values may
      be overridden on the command line or in an associated INI file
    }
    MyConfiguration.add(aseKeyConfigFilePath, '');
    MyConfiguration.add(aseKeyLogToSysLog, aseDefaultLogToSysLog);
    MyConfiguration.add(aseKeyLogFilePath, '');
    MyConfiguration.add(aseKeyEjectIdleAfterSeconds,
      aseDefaultEjectIdleAfterSeconds);

    MyIdleSessions := TQueue.create;
    if MyIdleSessions = nil then
      exit;

    result := true;
  end;
end;

// Destructor
destructor AnASEApplication.destroy;

begin
  if MyIdleSessions <> nil then
    SysUtils.freeAndNil(MyIdleSessions);

  if MyLog <> nil then
    SysUtils.freeAndNil(MyLog);

  if MyConfiguration <> nil then
    SysUtils.freeAndNil(MyConfiguration);

  inherited destroy;
end;

// Handle the occurrence of a command-line argument
{ If a command-line command was processed (an argument that begins with `-` or
  `--`, then `argument` will not be an empty string, and `parameter` may
  contain the parameter passed with the specified argument (e.g., as
  `command=parameter`).  Commands are passed in lower-case; parameters are
  passed as they were specified on the command line.

  If no command was processed, then `argument` will be an empty string and
  `parameter` will be the argument itself.
}
function AnASEApplication.handleCommandLineArgument(const argument: string;
  const parameter: string): boolean;

begin
  result := false;

  if argument <> '' then
    case argument of
      // Help with usage
      '?', 'help': begin
        Self.printUsage;
        exit;
      end;

      // Path to configuration file
      'cf', 'config':
        MyConfiguration[aseKeyConfigFilePath] :=
          ansiDequotedStr(Self.expandEnvironmentVariablesIn(parameter), '"');

      // Log to the named file
      'lf', 'log': begin
        MyConfiguration[aseKeyLogToSysLog] := 'n';
        MyConfiguration[aseKeyLogFilePath] :=
          ansiDequotedStr(Self.expandEnvironmentVariablesIn(parameter), '"');
      end;

      // Log to syslog
      'ls', 'syslog': begin
        MyConfiguration[aseKeyLogToSysLog] := 'y';
        MyConfiguration[aseKeyLogFilePath] := '';
      end;

      // Maximum number of seconds a disconnected session may be idle
      't', 'timeout':
        MyConfiguration[aseKeyEjectIdleAfterSeconds] :=
          ansiDequotedStr(parameter, '"');

      else begin
        Self.logErrorFmt(aseErrorUnrecognizedArgument, [argument]);
        exit;
      end;
    end;

  result := true;
end;

// Read configuration
function AnASEApplication.readConfiguration: boolean;

var
  // Stores the name of the configuration file
  configFilePath: string;
  // Used to process the configuration file
  ConfigFile: TIniFile;
  // Used to store the names of the sections in the configuration file
  ConfigSections: TStringList;
  // Used to store the names of keys and values in a section
  ConfigKeysValues: TStringList;
  // Used to store a configuration key name
  configKeyName: string;
  // Loop counters
  i, j: sizeInt;

begin
  result := false;

  configFilePath := MyConfiguration[aseKeyConfigFilePath];

  if configFilePath = '' then
  begin
    { No INI file path specified on command-line; fall back to the default
      INI file name, which should be located in the same directory as our
      executable, and have the same name as our executable
    }
    configFilePath :=
      Self.expandEnvironmentVariablesIn(
          SysUtils.format(aseFmtINIFilePath, [
            Self.executableName(true)
          ])
        );

    { It isn't an error if our default INI file is not found; it simply means
      we've either been configured from the command-line or we're falling back
      on our default options.
    }
    if SysUtils.fileExists(configFilePath) = false then
    begin
      result := true;
      exit;
    end;
  end;

  { If we make it here, an INI file path and filename were explicitly specified
    on the command line.  If the file is not found, then we presume the caller
    wants to know, since they took the trouble of specifying it.
  }
  if SysUtils.fileExists(configFilePath) = false then
  begin
    Self.logErrorFmt(aseErrorMissingINI, [configFilePath]);
    exit;
  end;

  try
    ConfigFile := nil;
    ConfigSections := nil;
    ConfigKeysValues := nil;

    // Open the config file
    ConfigFile := TIniFile.create(configFilePath);

    // Acquire a list of all sections in the file
    ConfigSections := TStringList.create;
    ConfigFile.readSections(ConfigSections);

    // Read the keys and values from each section
    ConfigKeysValues := TStringList.create;
    ConfigKeysValues.delimiter := '=';

    i := 0;
    while i < ConfigSections.count do
    begin
      ConfigFile.readSectionValues(ConfigSections[i], ConfigKeysValues);

      // Add each key/value pair to our configuration
      j := 0;
      while j < ConfigKeysValues.count do
      begin
        // Ensure that key case is irrelevant
        configKeyName := lowerCase(ConfigKeysValues.names[j]);

        { We silently ignore any unrecognized key names; a key is only
          considered recognized if it has already been added to MyConfiguration
        }
        if MyConfiguration.find(configKeyName) <> nil then
          MyConfiguration[configKeyName] := ConfigKeysValues.valueFromIndex[j];

        inc(j);
      end;

      ConfigKeysValues.clear;
      inc(i);
    end;

    result := SysUtils.tryStrToQWord(MyConfiguration[aseKeyEjectIdleAfterSeconds],
      mySessionIdleMaximum);

  finally
    if ConfigKeysValues <> nil then
      SysUtils.freeAndNil(ConfigKeysValues);

    if ConfigSections <> nil then
      SysUtils.freeAndNil(ConfigSections);

    if ConfigFile <> nil then
      SysUtils.freeAndNil(ConfigFile);
  end;
end;

// Begin logging
function AnASEApplication.beginLogging: boolean;

begin
  result := false;

  MyLog := TEventLog.create(nil);
  if MyLog = nil then
    exit;

  MyLog.identification := Self.title;

  if boolFromString(MyConfiguration[aseKeyLogToSyslog]) then
    { Log to system event log, if available.  If unavailable, the instance will
      log to a file named after the executable, which will typically be located
      in %TEMP%.
    }
    MyLog.logType := ltSystem

  else
    MyLog.logType := ltFile;

  { In the event we're logging to a file, or if the syslog is unavailable to us,
    we log to the file name specified in our configuration file
  }
  MyLog.fileName :=
    Self.expandEnvironmentVariablesIn(MyConfiguration[aseKeyLogFilePath]);

  result := true;
end;

// Calculate the idle time for a given session
function AnASEApplication.getIdleTimeSecondsFrom(const thisSession: PWTSINFO): qword;

const
  { Windows file times are stored as 100-nanosecond intervals.  We need to
    convert these intervals into seconds.
  }
  WINDOWS_FILETIME_TO_SECONDS = 10000000;

begin
  result := 0;

  if (thisSession = nil) or (thisSession^.State <> WTSDisconnected) then
    exit;

  // Determine how long the session has been disconnected
  result :=
    (thisSession^.CurrentTime.QuadPart - thisSession^.DisconnectTime.QuadPart) div
      WINDOWS_FILETIME_TO_SECONDS;
end;

// Collect idle sessions
function AnASEApplication.collectIdleSessions: sizeInt;

var
  // The query level
  queryLevel: DWORD;
  // The number of desktop sessions on the current machine
  desktopSessionCount: DWORD;
  // Refers to the list of enumerated desktop sessions
  desktopSession: PWTS_SESSION_INFO_1;
  // A loop counter
  i: DWORD;
  // Used when querying for additional information about each session
  sessionInfo: PWTSINFO;

begin
  result := -1;

  // Enumerate all sessions
  // Per MSDN, this must be one
  queryLevel := 1;
  // No desktop sessions enumerated yet
  desktopSessionCount := 0;
  // No desktop sessions enumerated yet
  desktopSession := nil;

  // Enumerate desktop sessions
  if WTSEnumerateSessionsEx(WTS_CURRENT_SERVER_HANDLE, @queryLevel, 0,
    @desktopSession, @desktopSessionCount) = false then
    // Failed to enumerate sessions
    exit;

  // We skip session 0 because it's a special session used by system services
  i := 1;
  // Loop through each session
  while i < desktopSessionCount do
  begin
    // We only care about disconnected sessions
    if desktopSession[i].State = WTSDisconnected then
    begin
      sessionInfo := nil;

      // Query for additional information about the session
      if WTSQuerySessionInformation(WTS_CURRENT_SERVER_HANDLE,
        desktopSession[i].SessionId, WTSSessionInfo, @sessionInfo,
        @queryLevel) then
      begin
        if Self.getIdleTimeSecondsFrom(sessionInfo) >= mySessionIdleMaximum then
          MyIdleSessions.push(sessionInfo)

        else
          WTSFreeMemory(sessionInfo);
      end;
    end;

    inc(i);
  end;

  WTSFreeMemoryEx(WTSTypeSessionInfoLevel1, desktopSession,
    desktopSessionCount);

  result := MyIdleSessions.count;
end;

// Eject idle sessions
function AnASEApplication.ejectIdleSessions: sizeInt;

var
  // The user names associated with idle sessions
  idleSessionUserNames: string;
  // Refers to an idle session
  thisIdleSession: PWTSINFO;

begin
  result := Self.collectIdleSessions;

  if result > 0 then
  begin
    idleSessionUserNames := '';

    thisIdleSession := PWTSINFO(MyIdleSessions.pop);
    while thisIdleSession <> nil do
    begin
      // Collect the name of the session user
      idleSessionUserNames := idleSessionUserNames +
        SysUtils.format(aseFmtIdleSessionUserName, [
          thisIdleSession^.SessionID,
          pchar(@thisIdleSession^.Domain[0]),
          pchar(@thisIdleSession^.UserName[0]),
          Self.getIdleTimeSecondsFrom(thisIdleSession)
        ]);

      // Log off the idle session
      WTSLogoffSession(WTS_CURRENT_SERVER_HANDLE, thisIdleSession^.SessionId,
        true);

      WTSFreeMemory(thisIdleSession);
      thisIdleSession := PWTSINFO(MyIdleSessions.pop);
    end;

    Self.logNoteFmt(aseNoteEjectIdleSessions, [
      mySessionIdleMaximum, idleSessionUserNames
    ]);
  end

  else if result = 0 then
    Self.logNoteFmt(aseNoteNoIdleSessions, [mySessionIdleMaximum])

  else
    Self.logErrorMessage(aseErrorCollectIdleSessions);
end;

// Execute the main body of the application
function AnASEApplication.run: ptrInt;

begin
  result := inherited run;
  if result < 0 then
    exit;

  result := -1;

  if Self.readConfiguration = false then
  begin
    Self.logErrorMessage(aseErrorReadingConfiguration);
    exit;
  end;

  if SysUtils.tryStrToQWord(MyConfiguration[aseKeyEjectIdleAfterSeconds],
    mySessionIdleMaximum) = false then
    exit;

  if (Self.beginLogging) and (Self.ejectIdleSessions >= 0) then
    result := 0;
end;

// Print usage information to stdout
procedure AnASEApplication.printUsage;

begin
  writeln(SysUtils.format(aseMsgUsageFmt, [
    Self.title, Self.VersionInfo[vinfFileVersion],
    SysUtils.extractFileName(Self.executableName(true)),
    aseDefaultEjectIdleAfterSeconds
  ]));
end;

// Log a note
procedure AnASEApplication.logNoteMessage(const thisMessage: string);

begin
  if MyLog <> nil then
    MyLog.info(thisMessage)

  else
    writeln(thisMessage);
end;

// Log a note
procedure AnASEApplication.logNoteFmt(const fmt: string;
  const args: array of const);

begin
  if MyLog <> nil then
    MyLog.info(fmt, args)

  else
    writeln(SysUtils.format(fmt, args));
end;

// Log an error message
procedure AnASEApplication.logErrorMessage(const thisMessage: string);

begin
  if MyLog <> nil then
    MyLog.error(thisMessage)

  else
    writeln(stderr, thisMessage);
end;

// Log an error message
procedure AnASEApplication.logErrorFmt(const fmt: string;
  const args: array of const);

begin
  if MyLog <> nil then
    MyLog.error(fmt, args)

  else
    writeln(stderr, SysUtils.format(fmt, args));
end;

end.

